export type IDSelectorStr<T> = (model: T) => string;
export type IDSelectorNum<T> = (model: T) => number;
export type IDSelector<T> = IDSelectorStr<T> | IDSelectorNum<T>;

export interface DictionaryNum<T> {
  [id: number]: T;
}
export abstract class Dictionary<T> implements DictionaryNum<T> {
  [id: string]: T;
}

export interface UpdateStr<T> {
  id: string;
  changes: Partial<T>;
}

export interface UpdateNum<T> {
  id: number;
  changes: Partial<T>;
}

export type Update<T> = UpdateStr<T> | UpdateNum<T>;

export interface EntityState<T, E> {
  ids: string[] | number[];
  entities: Dictionary<T>;
  error: E | null;
}

export interface EntityStateAdapter<T, E> {
  setOne<S extends EntityState<T, E>>(entity: T, state: S): S;
  setAll<S extends EntityState<T, E>>(entities: T[], state: S): S;

  addOne<S extends EntityState<T, E>>(entity: T, state: S): S;
  addMany<S extends EntityState<T, E>>(entities: T[], state: S): S;

  updateOne<S extends EntityState<T, E>>(entity: T, state: S): S;
  updateMany<S extends EntityState<T, E>>(entities: T[], state: S): S;

  upsertOne<S extends EntityState<T, E>>(entity: T, state: S): S;
  upsertMany<S extends EntityState<T, E>>(entities: T[], state: S): S;

  removeOne<S extends EntityState<T, E>>(entity: T, state: S): S;
  removeMany<S extends EntityState<T, E>>(entities: T[], state: S): S;
  removeAll<S extends EntityState<T, E>>(state: S): S;

  setError<S extends EntityState<T, E>>(error: E, state: S): S;
}

export interface EntitySelectors<T, E, S> {
  selectIds: (state: S) => string[] | number[];
  selectEntities: (state: S) => Dictionary<T>;
  selectOne: (id: string | number) => (state: S) => T;
  selectAll: (state: S) => T[];
  selectTotal: (state: S) => number;
  selectError: (state: S) => E;
}

export interface EntityAdapter<T, E> extends EntityStateAdapter<T, E> {
  selectId: IDSelector<T>;
  getInitialState(): EntityState<T, E>;
  getInitialState<S extends object>(state: S): EntityState<T, E> & S;
  getSelectors(): EntitySelectors<T, E, EntityState<T, E>>;
  getSelectors<S>(
    selectState: (state: S) => EntityState<T, E>
  ): EntitySelectors<T, E, S>;
}
